/**
 * \file  4dt_scn.c
 * \brief Modul for handling game scene
 */
 
/*------------------------------------------------------------------------------
   INCLUDE FILES
------------------------------------------------------------------------------*/

// Include standard library.
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "4dt_m4d.h"
#include "4dt_eng.h"
#include "4dt_g3d.h"
#include "4dt_g4d.h"
#include "4dt_menu.h"

/*------------------------------------------------------------------------------
   MACROS
------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------
   TYPES
------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------
   CONSTANTS
------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------
   GLOBAL VARIABLES
------------------------------------------------------------------------------*/

/** Color of the 4D cube. */
static float scn4DCubeColor[4] = {0.4, 0.4, 0.6, 0.15};

/** array of the colors of game space levels. */
static float scnLevelColors[SPACELENGTH][4];

/** selected axle for rotation */
int scnAxle = 0;

/*------------------------------------------------------------------------------
   PROTOTYPES
------------------------------------------------------------------------------*/

static void scnDrawBG(void);
static void scnWriteScore(void);
static void scnInitLevelColors(void);
static void scnDrawRotPlanes(void);

/*------------------------------------------------------------------------------
   FUNCTIONS
------------------------------------------------------------------------------*/

/** \brief Set random colors for game levels */
static void scnInitLevelColors(void)
{
  // Loop counters.
  int i, j;

  // For each level of the game space,
  for (i = 0; i < SPACELENGTH; i++)
  {
    // for each color component
    for (j = 0; j < 3; j++)
    {
      // create a random value.
      scnLevelColors[i][j] = (double)rand() / RAND_MAX;
    }
    // Set the color's alpha component.
    scnLevelColors[i][3] = 1.0;
  }
}

/** Write out the score to the game window. */
static void scnWriteScore(void)
{
  // Local variables:
  char text[30];     // buffer for the full text;
  float color[4] = {0.2, 0.2, 0.4, 0.8};

  // Create the score character array.
  sprintf(text, "Score: %i", engGE.score);

  // Render the text.
  g3dRenderString(0.1, 0.1, color, text);
}

/** Draw background for scene */
static void scnDrawBG(void)
{
  float color1[4] = {1.0, 1.0, 1.0, 1.0};
  float color2[4] = {0.8, 0.8, 0.8, 1.0};

  // Draw background
  g3dDrawRectangle(0.0, 0.0, 1.0, 1.0, color1, color2);
}

/** draws the rotation planes selected */
static void scnDrawRotPlanes(void)
{
  int i;
  const double planeSize = 3.5;

  float colors[3][4] = { {0.4, 0.4, 0.6, 0.8},
                         {0.4, 0.4, 0.6, 0.0},
                         {0.4, 0.4, 0.6, 0.0}};

  if ((scnAxle <= 2) && (scnAxle >= 0))
  {
    for (i = -1; i <= 1; i += 2)
    {
      float points[3][4] = { {0.0, 0.0, 0.0, 0.0},
                             {0.0, 0.0, 0.0, 0.0},
                             {0.0, 0.0, 0.0, 0.0}};

      points[1][scnAxle] = i * planeSize;
      points[2][scnAxle] = i * planeSize;

      g4dDrawTriangle(points, colors, 1);
    }
  }
}

/** Initialize the scene */
void scnInit(void)
{
  // Initialize random number generator with the
  // Answer to the Ultimate Question of Life, the Universe, and Everything
  // to get always the same colors for levels.
  srand(42);

  scnInitLevelColors();
}

/** Main drawing function. */
void scnDisplay(void)
{
  // Local variables:
  int n, l, x, y, z;        // loop counter;

  // mask indicates which block space
  // hidden by upper blocks
  int mask[2][2][2] = {{{0,0},{0,0}},{{0,0},{0,0}}};

  g3dBeginDraw();

  scnDrawBG();

  // Write out the game score.
  scnWriteScore();

  // Draw the gamespace.

  // For each level from top
  for (l = SPACELENGTH - 1; l >= 0; l--)
  {
    // For each cell of the level
    for (x = 0; x <= 1; x++)
    for (y = 0; y <= 1; y++)
    for (z = 0; z <= 1; z++)
    {
      // space which has no cube above (so it is visible)
      // gets rid of Z-fighting
      if (mask[x][y][z] == 0)
      {
        // if the cell is not empty then
        if (engGetSpaceCell(l, x, y, z))
        {
          // draw the cube.
          g4dDraw4DCube(m4dVector(x - 0.5, y - 0.5, z - 0.5, l - 0.5),
                        m4dUnitMatrix(),
                        scnLevelColors[l], 4, 0);

          mask[x][y][z] = 1;
        }
      }
    }
  }

  // Draw the bottom level.

  // For each cell of the level do:
  for (x = 0; x <= 1; x++)
  for (y = 0; y <= 1; y++)
  for (z = 0; z <= 1; z++)
  {
    // space which has no cube above (so it is visible)
    if (mask[x][y][z] == 0)
    {
      g4dDraw4DCube(m4dVector(x - 0.5, y - 0.5, z - 0.5, -0.5),
                    m4dUnitMatrix(),
                    scn4DCubeColor, 3, 1);
    }
  }

  // Draw the actual solid.

  // For each cell
  for (n = 0; n < engGE.object.block.num; n++)
  {
    tM4dVector pos;

    pos = m4dAddVectors(engGE.object.pos,
                        m4dMultiplyMV(engGE.object.axices,
                                      engGE.object.block.c[n]));

    // draw the hypercube.
    g4dDraw4DCube(pos, engGE.object.axices, scn4DCubeColor, 4, 1);
  }

  scnDrawRotPlanes();

  // draw the menu
  if (menuActived())
  {
    menuDraw();
  }

  g3dEndDraw();
}
